require 'installer/errors'
require 'net/ssh'
require 'net/scp'
require 'strscan'

class Installer::Host
  attr_reader :remote_conn, :roles, :hostname, :ssh_username, :ssh_key_file, :ssh_key_passphrase, :alt_names

  attr_accessor :data

  def initialize(hostname, config={})
    @hostname = hostname
    @alt_names = config[:alt_names] || []
    @roles = config[:roles] || []
    @puppet_managed = config[:puppet_managed] || true
    @ssh_username = config[:ssh_username] unless (config[:ssh_username].nil? || config[:ssh_username].strip.empty?)
    if config[:ssh_key_file] && ! config[:ssh_key_file].strip.empty?
      @ssh_key_file = File.expand_path(config[:ssh_key_file])
    end
    @ssh_key_passphrase = config[:ssh_key_passphrase] unless (config[:ssh_key_passphrase].nil? || config[:ssh_key_passphrase].strip.empty?)
    @sudo_password = config[:sudo_password] unless (config[:sudo_password].nil? || config[:sudo_password].empty?)
    @data = {}
  end

  def inbound
    @roles.collect { |r| r.inbound }.flatten.uniq
  end

  def outbound
    @roles.collect { |r| r.outbound }.flatten.uniq
  end

  def installed_services
    @roles.collect { |r| r.installed_services }.flatten.uniq
  end

  def assign_role(role)
    role.host = self
    @roles << role
  end

  def get_role_by_name(role_name)
    desired_role = nil
    @roles.each do |role|
      desired_role = role if role.name == role_name
    end
    desired_role
  end

  def requires_roles
    @roles.collect { |r| r.requires }.compact.flatten
  end

  def answers_hash
    defaults = {
      'q_install' => 'y',
      'q_puppet_cloud_install' => 'n',
      'q_vendor_packages_install' => 'y',
      'q_puppetmaster_install' => 'n',
      'q_puppetdb_install' => 'n',
      'q_puppet_enterpriseconsole_install' => 'n',
      'q_skip_backup' => 'y',
    }
    role_answers = @roles.inject({}) do |answers, role|
      answers.merge(role.answers_hash) do |k, v1, v2|
        raise Installer::Error, "Received conflicting answers for #{k}: #{v1.inspect} and #{v2.inspect} from #{role.name}" if v1 != v2
        v1
      end
    end

    defaults.merge!(role_answers)

  end

  # Initializes SSH connection/session to host. Must be called before {#ssh} or {#scp}.
  #
  # @returns [Net::SSH::Connection] SSH connection to host.
  def ssh_connect
    ssh_opts = {
      timeout: 30,
      passphrase: @ssh_key_passphrase,
      password: @sudo_password,
    }.reject { |k,v| v.nil? }

    ssh_opts[:keys] = [@ssh_key_file] if @ssh_key_file

    @ssh = Net::SSH.start(@hostname, @ssh_username, ssh_opts)
  rescue Net::SSH::AuthenticationFailed, Net::SSH::HostKeyError, Errno::ETIMEDOUT, SocketError, Timeout::Error => err
    raise(Installer::RemoteError, "Could not establish connection: #{err.message}")
  end

  # Shut down the SSH connection/session. Will block until all channels have closed. Removes session from host and returns the closed session.
  #
  # @returns [Net::SSH::Connection] Closed SSH connection
  def ssh_disconnect
    @ssh.close
    closed_session = @ssh
    @ssh = nil
    closed_session
  end

  # Returns the SSH session that connects to this host. {#ssh_connect} *must* be called before this will return a session, otherwise it will raise an exception.
  #
  # @returns [Net::SSH::Connection] SSH connection to host.
  def ssh
    @ssh || raise("SSH called but connection has not been established for #{@hostname}")
  end

  # If you would like to open an SCP channel and perform up uploads or downloads:
  # host.scp.download!(@from, @from_file.path, @opts, &stream_block)
  # host.scp.upload!(@from_file.path, @to, @opts, &stream_block)
  def scp
    ssh.scp
  end

  def upload!(*args, &blk)
    scp.upload!(*args, &blk)
  end

  def execute(cmd, sudo=false, &stream_block)
    results = {:exit_status => nil, :stdout => "", :stderr => ""}
    sudo = false if @ssh_username == 'root'

    if sudo
      # Because we're using a pty, and therefore a shell...
      cmd_no_quotes = cmd.delete("'")
      sudo_prompt = "[sudo] Puppet Enterprise Installer needs to run #{cmd_no_quotes} as root, please enter password: "
      cmd = "sudo -S -p '#{sudo_prompt}' #{cmd}"
    end

    channel = ssh.open_channel do |channel|
      channel.request_pty do |_, success|
        raise(Installer::RemoteError, "Could not acquire tty") unless success

        channel.exec cmd do |_, success|
          raise "could not execute command" unless success

          # Buffering so output doesn't get split up in strange ways
          buffers = {stdout: StringScanner.new(""), stderr: StringScanner.new("")}
          buffer_proc = Proc.new do |ch, stream, data|
            raise NotImplementedError, "Unknown stream #{stream}" if ![:stdout, :stderr].include? stream
            buffers[stream] << data
            while l = buffers[stream].scan(/[^\n]+\n/)
              stream_block.call(channel, stream, l)
            end
            buffers[stream].string = buffers[stream].rest
          end

          # "on_data" is called when the process writes something to stdout
          channel.on_data do |_, data|
            # For some reason, sudo prompt comes over stdout when we are using pty
            if sudo && sudo_stdin(channel, data, sudo_prompt, &stream_block)
              # Fake stderr
              buffer_proc.call(channel, :stderr, data)
              results[:stderr] << data
            else
              buffer_proc.call(channel, :stdout, data)
              results[:stdout] << data
            end
          end

          # "on_extended_data" is called when the process writes something to stderr
          channel.on_extended_data do |_, stream, data|
            if stream == 1
              stream = :stderr
            else
              raise NotImplementedError, "Unknown stream: #{stream}"
            end

            buffer_proc.call(channel, stream, data)
            results[stream] << data
          end

          channel.on_request("exit-status") do |_, data|
            results[:exit_status] = data.read_long
          end
        end
      end
    end
    channel.wait

    results
  end

  def generate_answers
    answers = []
    answers_hash.each do |q, a|
      answers << "#{q}='#{a}'"
    end
    answers.sort.join("\n")
  end

  def to_s
    @hostname
  end

  private

  def sudo_stdin(ch, data, sudo_prompt, &stream_block)
    # Sudo handling
    if data == sudo_prompt
      if @sudo_password
        ch.send_data "#{@sudo_password}\r\n"
      else
        raise Installer::RemoteError, "Sudo password for the SSH user #{@ssh_username} was not provided"
      end
      ch.wait

      true
    elsif data =~ /^#{@ssh_username} is not in the sudoers file./
      # Sudo failed, wrong user. Bail out.
      stream_block.call(ch, :stderr, "Cannot proceed: SSH User #{@ssh_username} does not have sudo permission.")
      raise Installer::RemoteError, "SSH User #{@ssh_username} does not have sudo permission"
    # This could be a terrible bug.
    elsif data =~ /Sorry, try again./
      # Sudo failed, wrong password. Bail out.
      stream_block.call(ch, :stderr, "Cannot proceed: Sudo password not recognized.")
      raise Installer::RemoteError, "Sudo password not recognized"
    end
  end
end
